/**
 * 
 */
package analyzer;

import java.io.*;
import java.text.Normalizer;
import java.util.*;
import java.util.ArrayList;

import org.tartarus.snowball.SnowballStemmer;
import org.tartarus.snowball.ext.englishStemmer;
import org.tartarus.snowball.ext.porterStemmer;

import json.JSONArray;
import json.JSONException;
import json.JSONObject;
import opennlp.tools.tokenize.Tokenizer;
import opennlp.tools.tokenize.TokenizerME;
import opennlp.tools.tokenize.TokenizerModel;

import structures.*;
import utils.*;

import javax.rmi.CORBA.Util;

/**
 * @author hongning
 * Sample codes for demonstrating OpenNLP package usage 
 * NOTE: the code here is only for demonstration purpose, 
 * please revise it accordingly to maximize your implementation's efficiency!
 */
public class DocAnalyzer {
	//N-gram to be created
	int m_N;
	
	//a list of stopwords
	HashSet<String> m_stopwords;
	
	//you can store the loaded reviews in this arraylist for further processing
	ArrayList<Post> m_reviews;
	
	//feature stats, such as DF
	HashMap<String, Token> m_stats;

    // controlled vocabulary
	HashMap<String, Integer> m_CV;

	//number of doc for each class
	int[] m_classStats;

	boolean m_isCVLoaded = false;
	int m_lengthThreshold = 5;

	int m_hashNum;
	KNN m_knn;

	//we have also provided a sample implementation of language model in src.structures.LanguageModel
	Tokenizer m_tokenizer;
	
	public DocAnalyzer(String tokenModel, int N) throws IOException {
		m_N = N;
		m_reviews = new ArrayList<Post>();
		m_tokenizer = new TokenizerME(new TokenizerModel(new FileInputStream(tokenModel)));
        m_stats = new HashMap<String, Token>();
        m_CV = new HashMap<>();
        m_stopwords = new HashSet<>();
	}

	public DocAnalyzer(String tokenModel, int Ngram, int classNum, int minDocLength, int hashnum) throws IOException {
        m_reviews = new ArrayList<Post>();
        m_tokenizer = new TokenizerME(new TokenizerModel(new FileInputStream(tokenModel)));
        m_stopwords = new HashSet<>();

        m_N = Ngram;
        m_classStats = new int[classNum];
        m_hashNum = hashnum;
        m_stats = new HashMap<String, Token>();
        m_lengthThreshold = minDocLength;

        m_CV = new HashMap<>();
    }

    public ArrayList<Post> getReviews() { return m_reviews; }

    public HashMap<String, Token> getTokenStats(){ return m_stats; }

    public int[] getClassStats() { return m_classStats; }

    public HashMap<String, Integer> getCV() { return m_CV; }

	//sample code for loading a list of stopwords from file
	//you can manually modify the stopword file to include your newly selected words
	public void LoadStopwords(String filename) {
		try {
			BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));
			String line;

			int size = m_stopwords.size();
			while ((line = reader.readLine()) != null) {
				//it is very important that you perform the same processing operation to the loaded stopwords
				//otherwise it won't be matched in the text content
				line = SnowballStemming(Normalization(line));
				if (!line.isEmpty())
					m_stopwords.add(line);
			}
			reader.close();
			System.out.format("Loading %d stopwords from %s\n", m_stopwords.size() - size, filename);
		} catch(IOException e){
			System.err.format("[Error]Failed to open file %s!!", filename);
		}
	}

	public boolean LoadCV(String filename){
	    if (filename == null || filename.isEmpty())
	        return false;
        try {
            BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));
            String line;
            int idx = 0;
            while ((line = reader.readLine()) != null) {
                String word = line.split("\t")[0];
                m_CV.put(word, idx++);
            }
            reader.close();

            System.out.format("%d feature words loaded from %s...\n", m_CV.size(), filename);
            m_isCVLoaded = true;

            return true;
        } catch (IOException e) {
            System.err.format("[Error]Failed to open file %s!!", filename);
            return false;
        }
    }

	public void LoadStats(String filename, HashMap<String, Token> stats){
	    try{
	        BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "utf-8"));
	        String line;

	        while((line = reader.readLine()) != null){
                if(!line.isEmpty()) {
                    String[] twoParts = line.split("\t");
                    Token token = new Token(twoParts[0]);
                    token.setValues(0, Double.parseDouble(twoParts[1]));
                    stats.put(twoParts[0], token);
                }
            }
            reader.close();
            System.out.format("Loading %d stats from %s\n", stats.size(), filename);
        } catch (IOException e){
	        System.err.format("[Error]Failed to open stats file %s!!", filename);
        }
    }

    public boolean isWord(String token){
        return !token.isEmpty()
//                && !m_stopwords.contains(token)
                && token.length()>1
                && token.length()<20;
    }

    public boolean isLegit(String token){
        return !token.isEmpty()
                && !m_stopwords.contains(token)
                && token.length()>1
                && token.length()<40;
    }

    public boolean isFeature(String token){
        return !token.isEmpty()
                && !m_stopwords.contains(token)
                && m_stats.containsKey(token)
                && token.length()>1
                && token.length()<40;
    }

    // sample code for demonstrating how to recursively load files in a directory
    public void representDocuments(String folder, String suffix, String type) {
        File dir = new File(folder);
        int size = m_reviews.size();
        for (File f : dir.listFiles()) {
            if (f.isFile() && f.getName().endsWith(suffix))
                analyzeDocument(LoadJson(f.getAbsolutePath()), "FIX", type);
            else if (f.isDirectory())
                representDocuments(f.getAbsolutePath(), suffix, type);
        }
        size = m_reviews.size() - size;
        System.out.println("Load " + size + " review documents from " + folder);
    }

    // sample code for demonstrating how to recursively load files in a directory
    public void LoadDirectory(String folder, String suffix, String type) {
        File dir = new File(folder);
        int size = m_reviews.size();
        for (File f : dir.listFiles()) {
            if (f.isFile() && f.getName().endsWith(suffix))
                analyzeDocument(LoadJson(f.getAbsolutePath()), "STATS", type);
            else if (f.isDirectory())
                LoadDirectory(f.getAbsolutePath(), suffix, type);
        }
        size = m_reviews.size() - size;
        System.out.println("Load " + size + " review documents from " + folder);
    }

	public void analyzeDocument(JSONObject json, String mode, String type) {
		try {
			JSONArray jarray = json.getJSONArray("Reviews");
			for(int i=0; i<jarray.length(); i++) {
				Post review = new Post(jarray.getJSONObject(i));
                String[] initTokens = Tokenize(review.getContent());
                double rate = review.getRating();
                int label = 0;
                if(rate >= 4)
                    label = 1;

                // normalization and stemming for Ngram
                LinkedList<String> Ngrams = new LinkedList<>();
                for(int n = 0; n < initTokens.length; n++){
                    initTokens[n] = SnowballStemming(Normalization(initTokens[n]));
                }
                for(int n = 0; n < initTokens.length; n++){
                    String token = initTokens[n];
                    boolean legit = isLegit(token);
                    if (legit)
                        Ngrams.add(token);

                    if (!token.isEmpty()) {
                        for (int j = n - 1; j >= Math.max(0, n - m_N + 1); j--) {
                            if (token.isEmpty()) {
                                break;//touch the boundary
                            }

                            token = initTokens[j] + "-" + token;
                            legit &= isLegit(initTokens[j]);
                        }

                        if (legit && token.contains("-")) {//they are all legitimate
                            Ngrams.add(token);
                        }
                    }
                }
                String[] tokens =Ngrams.toArray(new String[Ngrams.size()]);

                //construct sparse vector with different class label
                HashMap<String,Token> sparseVct = new HashMap<>();
                for(String token : tokens) {
                    if (mode.equals("STATS")) {
                        if (!m_isCVLoaded || m_CV.containsKey(token)) { // generate CV
                            if (!sparseVct.containsKey(token)) {
                                addToken(m_stats, token, label); // because m_stats is DF, only count once per doc
                            }
                            addToken(sparseVct, token, label);
                        }
                    } else {
                        if (m_stats.containsKey(token)) {
                            setToken(sparseVct, token, m_stats.get(token).getValue(0));
                        }
                    }
                }

                if(!m_isCVLoaded || sparseVct.size() > m_lengthThreshold){
                    review.setVct(sparseVct);
                    review.setType(type);
                    m_reviews.add(review);
                    m_classStats[label]++;
                }
			}
		} catch (JSONException e) {
			e.printStackTrace();
		}
	}

	// calculate TFIDF to replace DF in m_stats
    public void calcTFIDF(){
        HashMap<String, Token> tfStats = new HashMap<String, Token>();
	    for(Post review : m_reviews){
            HashMap<String,Token> vct = review.getVct();
            for(Map.Entry<String, Token> entry : vct.entrySet()){
                double tf = Utils.sumOfArray(entry.getValue().getValues()); // sparse vector in review is stored as TF
                addToken(tfStats, entry.getKey(), tf);

            }
        }
	    for (Map.Entry<String, Token> entry : m_stats.entrySet()) {
            double tf = Utils.sumOfArray(tfStats.get(entry.getKey()).getValues()); // in overall m_stats is stored as DF
            double idf = Utils.sumOfArray(entry.getValue().getValues()); // sparse vector in review is stored as TF
            if(tf > 0.0){
                double tfidf = (1.0 + Math.log(tf)) * idf;
                entry.getValue().setValues(0, tfidf);//store tfidf in m_values[0]
            }
        }
    }

    public void setToken(HashMap<String, Token> vector, String token, double value){ //for token without class, store stats in m_values[0]
        if(! vector.containsKey(token)){
            Token newTk = new Token(token);
            newTk.setValues(0, value);
            vector.put(token, newTk);
        }
    }

    public void addToken(HashMap<String, Token> vector, String token){ //for token without class, store stats in m_values[0]
        if(vector.containsKey(token)){
            vector.get(token).addValues(0);
        } else{
            Token newTk = new Token(token);
            newTk.addValues(0);
            vector.put(token, newTk);
        }
    }

    public void addToken(HashMap<String, Token> vector, String token, double value){ //for token without class, store stats in m_values[0]
        if(vector.containsKey(token)){
            vector.get(token).setValues(0, value);
        } else{
            Token newTk = new Token(token);
            newTk.setValues(0, value);
            vector.put(token, newTk);
        }
    }

    public void addToken(HashMap<String, Token> vector, String token, int index){
	    if(vector.containsKey(token)){
	        vector.get(token).addValues(index);
        } else{
	        Token newTK = new Token(token, m_classStats.length);
	        newTK.addValues(index);
	        vector.put(token, newTK);
        }
    }

	//sample code for loading a json file
	public JSONObject LoadJson(String filename) {
		try {
			BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));
			StringBuffer buffer = new StringBuffer(1024);
			String line;

			while((line=reader.readLine())!=null) {
				buffer.append(line);
			}
			reader.close();

			return new JSONObject(buffer.toString());
		} catch (IOException e) {
			System.err.format("[Error]Failed to open file %s!", filename);
			e.printStackTrace();
			return null;
		} catch (JSONException e) {
			System.err.format("[Error]Failed to parse json file %s!", filename);
			e.printStackTrace();
			return null;
		}
	}

	//sample code for demonstrating how to use Snowball stemmer
	public String SnowballStemming(String token) {
		SnowballStemmer stemmer = new englishStemmer();
		stemmer.setCurrent(token);
		if (stemmer.stem())
			return stemmer.getCurrent();
		else
			return token;
	}
	
	//sample code for demonstrating how to use Porter stemmer
	public String PorterStemming(String token) {
		porterStemmer stemmer = new porterStemmer();
		stemmer.setCurrent(token);
		if (stemmer.stem())
			return stemmer.getCurrent();
		else
			return token;
	}
	
	//sample code for demonstrating how to perform text normalization
	//you should implement your own normalization procedure here
	public String Normalization(String token) {
		// remove all non-word characters
		// please change this to removing all English punctuation
        token = Normalizer.normalize(token, Normalizer.Form.NFKC);
		token = token.replaceAll("\\W+", "");
		
		// convert to lower case
		token = token.toLowerCase();

		// add a line to recognize integers and doubles via regular expression
		// and convert the recognized integers and doubles to a special symbol "NUM"
        // consider case that: 123abc, so we use matcher to fully match a token
        if(token.matches("\\d+")){
            return "NUM";
        }
		
		return token;
	}
	
	String[] Tokenize(String text) {
		return m_tokenizer.tokenize(text);
	}

	public HashMap<String, Double> selectFeature(String model, int minDF, int maxDF){
	    FeatureSelector selector = new FeatureSelector(minDF, maxDF);
	    System.out.println("selecting features ...");
	    if(model.equals("IG")){
	        selector.IG(m_stats, m_classStats);
        }else if(model.equals("CHI")){
	        selector.CHI(m_stats, m_classStats);
        }
        return selector.getSelectedFeatures();
    }

}
